%% Script for trying some models on coupled EEM NMR LC-MS data

%------------------------------EvrimDataTestsv2-----------------------------%
% This file tests the Coupled CP ALS algorithms on Data by Acar et al.
%[E. Acar, E. E. Papalexakis, G. Gurdeniz,  M. A. Rasmussen,  A. J. Lawaetz,...
% M. Nilsson,  R. Bro, Structure-Revealing Data Fusion, BMC Bioinformatics, 15: 239, 2014.]
%
% Coments : here we implement the coupling of only some factors

% External functions:
%
% - kr.m              : Khatri-Rao product;
% - als.m             : ALS algorithm;
% - amb_correct_1f.m  : Matching Permutation for two CP models based on 
%                       factors C only;
% - amb_correct.m     : Matching Permutation for two CP models;
% - ccp_als.m         : Coupled CP ALS.
% - als_exact_c.m     : Exact couplec ALS
%
% List of updates                 -     04/02/2016  -     J.E. Cohen 
%                                       Creation of the file
%                                       08/05/2016  -     J.E. Cohen
%                                       Added exact coupling comparaison
%-------------------------------------------------------------------------%

%--------------Cleaning the workspace and loading functions---------------%
clc;
clear     all;
close     all;
%-------------------------------------------------------------------------%

load('/Users/cohenjer/Documents/Travail/MATLAB/Data/Data Chemo/data/Multimodal/EEM_NMR_LCMS.mat')
%load('D:/Travail/Travail/MATLAB/Data/Data Chemo/data/Multimodal/EEM_NMR_LCMS.mat')

%------------------------ Missing values preprocessing -------------------%

Tx        =    X.data;
Ty        =    Y.data;

Tx        =    permute(Tx,[3,2,1]);
Ty        =    permute(Ty,[3,2,1]);

% 3-way block dimensions
dim_X     =     size(Tx);
dim_Y     =     size(Ty);

Tx        =    reshape(Tx,prod(dim_X),1);
Ty        =    reshape(Ty,prod(dim_Y),1);

temp_x         =     Tx(isnan(Tx));
temp_y         =     Ty(isnan(Ty));

Tx(isnan(Tx))  =     0;%abs(0.01*randn(size(temp_x)));
Ty(isnan(Ty))  =     0;%abs(0.01*randn(size(temp_y)));

Tx             =     reshape(Tx,dim_X);
Ty             =     reshape(Ty,dim_Y);

normTx         =     (Tx(:))'*(Tx(:));
normTy         =     (Ty(:))'*(Ty(:));
Tx             =     Tx/sqrt(normTx); 
Ty             =     Ty/sqrt(normTy);
%-----------------------------Model parameters----------------------------%
% Number of components
R_X       =     3;
R_Y       =     5;
% Shared components
r         =     3;
% Standard deviations of the measurements noise
%sigma_n   =     [1 0.1];
sigma_n   =     [1 1];
% Standard deviation of the coupling
sigma_c   =     5;
% Number of iterations
iter      =     3000;
iter_j    =     500;
% Compression for the joint algorithm
compr     =     'yes';
% Number of initialization
N    =    10;

%-------------------------------------------------------------------------%

%----------------------- Tucker3 Coupled Compression ---------------------%

Tx_m      =     reshape( permute( Tx , [3,1,2] ) , dim_X(3) , dim_X(2)*dim_X(1) );
Ty_m      =     reshape( permute( Ty , [3,1,2] ) , dim_Y(3) , dim_Y(2)*dim_Y(1) );

dim1c     =     [4,4,6];%
dim2c     =     [6,6,6];%


[G_ind,U,V,W_ind,~,~,Sigma_1]     =     hosvd2(Tx,dim1c,'random');
[G_p_ind,U_p,V_p,W_p_ind,huhu,huhui,Sigma_2] =     hosvd2(Ty,dim2c,'random');

[W,Sigma_j,~]   = rsvd([Tx_m/sigma_n(1),Ty_m/sigma_n(2)],dim1c(3),1);
%[W,Sigma_j,~]   = svd([Tx_m/sigma_n(1),Ty_m/sigma_n(2)],'econ');
W_p       = W;

Gx   = MultProd(Tx,U',V',W');
Gy   = MultProd(Ty,U_p',V_p',W');

% Structure containing compression info for ccp_als
if strcmp(compr,'yes')
M{1} =      U;   M{2} =      V;    M{3} =      W;
M{4} =      U_p;   M{5} =      V_p;
end

%figure, hold on, plot(diag(Sigma_1),'b'),plot(diag(Sigma_2),'r'),plot(diag(Sigma_j),'m')
figure, hold on, plot(10*log10(diag(Sigma_1)),'b'),plot(10*log10(diag(Sigma_2)),'r'),plot(10*log10(diag(Sigma_j)),'m')
legend('\Sigma_1','\Sigma_2','\Sigma_j')


%-----------------------------Coupling model------------------------------%
% Constraint matrices for vectorized factors
% First tensor
if strcmp(compr,'yes')
H   =     [eye(r*dim1c(3)), zeros(r*dim1c(3),dim1c(3)*(R_X-r))];
else
H   =     [eye(r*dim_X(3)), zeros(r*dim_X(3),dim_X(3)*(R_X-r))];
end
% Second tensor
if strcmp(compr,'yes')
H_p =     [eye(r*dim2c(3)), zeros(r*dim2c(3),dim2c(3)*(R_Y-r))];
else
H_p =     [eye(r*dim_Y(3)), zeros(r*dim_Y(3),dim_Y(3)*(R_Y-r))];
end

%% 
%-------------------------------------------------------------------------%
%------------------------------Computation--------------------------------%


Stockerr1   =     inf;
Stockerr2   =     inf;
Stockerr3   =     inf;
Stockerr4   =     inf;
Stockerr5   =     inf;
Stockerr6   =     inf;

for n=1:N

%-----------------------------Initialization------------------------------%
% First tensor
A_0         =     abs(randn(dim_X(1),R_X));
B_0         =     abs(randn(dim_X(2),R_X));
A_0         =     A_0.*repmat(1./sqrt(sum(A_0.^2)),dim_X(1),1);
B_0         =     B_0.*repmat(1./sqrt(sum(B_0.^2)),dim_X(2),1);
% Second tensor   
A_p_0       =     abs(randn(dim_Y(1),R_Y));
B_p_0       =     abs(randn(dim_Y(2),R_Y));
A_p_0       =     A_p_0.*repmat(1./sqrt(sum(A_p_0.^2)),dim_Y(1),1);
B_p_0       =     B_p_0.*repmat(1./sqrt(sum(B_p_0.^2)),dim_Y(2),1);
 
C_0         =     abs(randn(dim_X(3),R_X));
C_p_0       =     abs(randn(dim_Y(3),R_Y));
%-------------------------------------------------------------------------%




%---------------------------Uncoupled ALS---------------------------------%
% Uncoupled ALS for the first tensor
[A_u,B_u,C_u,err_vec_1]     =     proco_als(Tx,iter,dim1c,'standard','given',A_0,B_0,C_0,0,Gx,U,V,W);
%[A_u,B_u,C_u,err_vec_1]     =     proco_als(Tx,iter,dim1c,'standard','given',A_0,B_0,C_0,0,G_ind,U,V,W_ind);
% ncoupled ALS for the second tensor
[A_p_u,B_p_u,C_p_u,err_vec_2]=    proco_als(Ty,iter,dim2c,'standard','given',A_p_0,B_p_0,C_p_0,0,Gy,U_p,V_p,W);
%[A_p_u,B_p_u,C_p_u,err_vec_2]=    proco_als(Ty,iter,dim2c,'standard','given',A_p_0,B_p_0,C_p_0,0,G_p_ind,U_p,V_p,W_p_ind);
                                    
% Correcting permutation for the initialization of the coupled algorithms

[A_p_u,B_p_u,C_p_u,A_u,B_u,C_u]   =           amb_correct_1f(A_p_u,B_p_u,C_p_u,A_u,B_u,C_u);

%-------------------------------------------------------------------------%
% 
%--------------Initialization with uncoupled results----------------------%
% Compressed versions
% First CP model
A_0 =     A_u;
B_0 =     B_u;
C_0 =     C_u;
% Second CP model
A_p_0     =    A_p_u;
B_p_0     =    B_p_u;
C_p_0     =    C_p_u;



%------------CCP ALS updated version (Exact version)------------%
tic
% CCP_ALS uncompressed
if strcmp(compr,'no')
[A_c_d,B_c_d,C_c_d,A_p_c_d,B_p_c_d,C_p_c_d,map_d_vec]  = ...
 ccp_als(Tx,Ty,iter_j,H,H_p,sigma_n,sigma_c,'joint',A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,'NN');
[A_ex,B_ex,C_ex,A_p_ex,B_p_ex,C_p_ex,err_als_vec_ex]=     als_exact(Tx,Ty,iter_j,A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,1,1,3);
else
% CCP ALS compressed
[A_c_d,B_c_d,C_c_d,A_p_c_d,B_p_c_d,C_p_c_d,map_d_vec]  = ...
 ccp_als(Gx,Gy,iter_j,H,H_p,sigma_n,sigma_c,'joint',A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,'NN',compr,M);
[A_c_d2,B_c_d2,C_c_d2,A_p_c_d2,B_p_c_d2,C_p_c_d2,map_d_vec2]  = ...
 ccp_als(Gx,Gy,iter_j,H,H_p,sigma_n,2,'joint',A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,'NN',compr,M);
[A_c_d3,B_c_d3,C_c_d3,A_p_c_d3,B_p_c_d3,C_p_c_d3,map_d_vec3]  = ...
 ccp_als(Gx,Gy,iter_j,H,H_p,sigma_n,0.5,'joint',A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,'NN',compr,M);

[A_ex,B_ex,C_ex,A_p_ex,B_p_ex,C_p_ex,err_als_vec_ex] =     als_exact_c(Gx,Gy,iter_j,A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,1,1,3,M);
end
toc

% Error Comparaison and keeping best run
if map_d_vec(end)<Stockerr1
    A_j     =    A_c_d;B_j    =    B_c_d;C_j    =    C_c_d;
    A_p_j   =    A_p_c_d;B_p_j=    B_p_c_d;C_p_j=    C_p_c_d;
    Stockerr1=map_d_vec(end);
end

if map_d_vec2(end)<Stockerr5
    A_j2     =    A_c_d2;B_j2    =    B_c_d2;C_j2    =    C_c_d2;
    A_p_j2   =    A_p_c_d2;B_p_j2=    B_p_c_d2;C_p_j2=    C_p_c_d2;
    Stockerr5=map_d_vec2(end);
end

if map_d_vec3(end)<Stockerr6
    A_j3     =    A_c_d3;B_j3    =    B_c_d3;C_j3    =    C_c_d3;
    A_p_j3   =    A_p_c_d3;B_p_j3=    B_p_c_d;C_p_j3=    C_p_c_d3;
    Stockerr6=map_d_vec3(end);
end

if err_als_vec_ex(end)<Stockerr4
    A_jex   =    A_ex;B_jex    =    B_ex;C_jex    =    C_ex;
    A_p_jex =    A_p_ex;B_p_jex=    B_p_ex;C_p_jex=    C_p_ex;
    Stockerr4=err_als_vec_ex(end);
end


if err_vec_1(end)<Stockerr2
    A_e     =    A_u;B_e    =    B_u;C_e    =    C_u;
    Stockerr2=err_vec_1(end);
end

if err_vec_2(end)<Stockerr3
    A_p_e   =    A_p_u;B_p_e  =    B_p_u;C_p_e  =    C_p_u;
    Stockerr3=err_vec_2(end);
end

end
%[A_p_e,B_p_e,C_p_e,A_e,B_e,C_e]   =           amb_correct_1f(A_p_e,B_p_e,C_p_e,A_e,B_e,C_e);
[A_p_e,B_p_e,C_p_e]         =           amb_correct(A_p_e,B_p_e,C_p_e,A_p_j,B_p_j,C_p_j);
[A_e,B_e,C_e]               =           amb_correct(A_e,B_e,C_e,A_j,B_j,C_j);
[A_jex,B_jex,C_jex]         =           amb_correct(A_jex,B_jex,C_jex,A_j,B_j,C_j);
[A_p_jex,B_p_jex,C_p_jex]   =           amb_correct(A_p_jex,B_p_jex,C_p_jex,A_p_j,B_p_j,C_p_j);
[A_j2,B_j2,C_j2]         =           amb_correct(A_j2,B_j2,C_j2,A_j,B_j,C_j);
[A_p_j2,B_p_j2,C_p_j2]   =           amb_correct(A_p_j2,B_p_j2,C_p_j2,A_p_j,B_p_j,C_p_j);
[A_j3,B_j3,C_j3]         =           amb_correct(A_j3,B_j3,C_j3,A_j,B_j,C_j);
[A_p_j3,B_p_j3,C_p_j3]   =           amb_correct(A_p_j3,B_p_j3,C_p_j3,A_p_j,B_p_j,C_p_j);
%% -------------------------------------------------------------------------%
close all

A_e     =     A_e.*repmat(1./sqrt(sum(A_e.^2)),dim_X(1),1);
A_p_e   =     A_p_e.*repmat(1./sqrt(sum(A_p_e.^2)),dim_Y(1),1);
A_j     =     A_j.*repmat(1./sqrt(sum(A_j.^2)),dim_X(1),1);
A_p_j   =     A_p_j.*repmat(1./sqrt(sum(A_p_j.^2)),dim_Y(1),1);
A_j2    =     A_j2.*repmat(1./sqrt(sum(A_j2.^2)),dim_X(1),1);
A_p_j2  =     A_p_j2.*repmat(1./sqrt(sum(A_p_j2.^2)),dim_Y(1),1);
A_j3    =     A_j3.*repmat(1./sqrt(sum(A_j3.^2)),dim_X(1),1);
A_p_j3  =     A_p_j3.*repmat(1./sqrt(sum(A_p_j3.^2)),dim_Y(1),1);
A_jex   =     A_jex.*repmat(1./sqrt(sum(A_jex.^2)),dim_X(1),1);
A_p_jex =     A_p_jex.*repmat(1./sqrt(sum(A_p_jex.^2)),dim_Y(1),1);

% Real Concentration profiles
ctrueval=[5,0,0,0,0,1.25,3.75,2.50,5,3.75,6.25,1.25,2.50,5,1.25,3.75,2.50,1.25,3.75,2.50,5,3.75,5,1.25,2.50,5,3.75,2.50];
ctruegly=[0,5,0,0,0,5,1.25,5,3.75,3.75,1.25,5,6.25,1.25,3.75,3.75,3.75,0,0,0,0,0,0,0,0,0,0,0];
ctruephe=[0,0,5,0,0,3.75,5,2.50,2.50,5,1.25,2.50,2.50,3.75,2.50,1.25,1.25,5,2.50,3.75,1.25,3.75,1.25,5,2.50,1.25,5,2.50];
ctruemal=[0,0,0,5,0,3.75,1.25,1.25,3.75,0,0,0,0,1.25,2.50,5,1.25,2.50,5,1.25,3.75,0,0,0,0,3.75,2.50,1.25];
ctruepro=[0,0,0,0,5,0,0,0,0,1.25,2.50,5,2.50,3.75,2.50,1.25,6.25,0,0,0,0,5,3.75,2.50,1.25,3.75,1.25,5];

ctrueval=ctrueval/sqrt(sum(ctrueval.^2));
ctruegly=ctruegly/sqrt(sum(ctruegly.^2));
ctruephe=ctruephe/sqrt(sum(ctruephe.^2));
ctruemal=ctruemal/sqrt(sum(ctruemal.^2));
ctruepro=ctruepro/sqrt(sum(ctruepro.^2));

Ctrue=[ctruephe',ctruegly',ctrueval',ctruemal',ctruepro'];
mal = 4;
pro = 5;

figure
subplot(231)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(A_e)
subplot(232)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(A_j))
subplot(233)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(A_jex)
subplot(234)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(A_p_e)
subplot(235)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(A_p_j))
subplot(236)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(A_p_jex))


figure
subplot(231)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(B_e)
subplot(232)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(B_j))
subplot(233)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(B_jex))
subplot(234)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(B_p_e)
subplot(235)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(B_p_j))
subplot(236)
set(gca,'ColorOrder', [0 0 1; 0 1 0; 1 0 0; 0 0.5 0.5; 0.5 0 0.5; 0.5 0.5 0; 0.5 0.5 0.5],'NextPlot', 'replacechildren')
plot(abs(B_p_jex))

figure
subplot(5,5,1)
hold on, plot(Ctrue(:,1),'k'), plot(C_e(:,1),'b'),plot(C_p_e(:,1),'r')
subplot(5,5,6)
hold on, plot(Ctrue(:,2),'k'), plot(C_e(:,2),'b'),plot(C_p_e(:,2),'r')
subplot(5,5,11)
hold on, plot(Ctrue(:,3),'k'), plot(C_e(:,3),'b'),plot(C_p_e(:,3),'r')
subplot(5,5,16)
hold on, plot(Ctrue(:,4),'k'), plot(C_p_e(:,mal),'r')
subplot(5,5,21)
hold on, plot(Ctrue(:,5),'k'), plot(C_p_e(:,pro),'r')

subplot(552)
hold on, plot(Ctrue(:,1),'k'), plot(C_j(:,1),'b'),plot(C_p_j(:,1),'r')
subplot(557)
hold on, plot(Ctrue(:,2),'k'), plot(C_j(:,2),'b'),plot(C_p_j(:,2),'r')
subplot(5,5,12)
hold on, plot(Ctrue(:,3),'k'), plot(C_j(:,3),'b'),plot(C_p_j(:,3),'r')
subplot(5,5,17)
hold on, plot(ctruemal,'k'), plot(C_p_j(:,mal),'r')
subplot(5,5,22)
hold on, plot(ctruepro,'k'), plot(C_p_j(:,pro),'r')

subplot(555)
hold on, plot(Ctrue(:,1),'k'), plot(C_jex(:,1),'b'),plot(C_p_jex(:,1),'r')
subplot(5,5,10)
hold on, plot(Ctrue(:,2),'k'), plot(C_jex(:,2),'b'),plot(C_p_jex(:,2),'r')
subplot(5,5,15)
hold on, plot(Ctrue(:,3),'k'), plot(C_jex(:,3),'b'),plot(C_p_jex(:,3),'r')
subplot(5,5,20)
hold on, plot(ctruemal,'k'), plot(C_p_jex(:,mal),'r')
subplot(5,5,25)
hold on, plot(ctruepro,'k'), plot(C_p_jex(:,pro),'r')

subplot(553)
hold on, plot(Ctrue(:,1),'k'), plot(C_j2(:,1),'b'),plot(C_p_j2(:,1),'r')
subplot(558)
hold on, plot(Ctrue(:,2),'k'), plot(C_j2(:,2),'b'),plot(C_p_j2(:,2),'r')
subplot(5,5,13)
hold on, plot(Ctrue(:,3),'k'), plot(C_j2(:,3),'b'),plot(C_p_j2(:,3),'r')
subplot(5,5,18)
hold on, plot(ctruemal,'k'), plot(C_p_j2(:,mal),'r')
subplot(5,5,23)
hold on, plot(ctruepro,'k'), plot(C_p_j2(:,pro),'r')

subplot(554)
hold on, plot(Ctrue(:,1),'k'), plot(C_j3(:,1),'b'),plot(C_p_j3(:,1),'r')
subplot(5,5,9)
hold on, plot(Ctrue(:,2),'k'), plot(C_j3(:,2),'b'),plot(C_p_j3(:,2),'r')
subplot(5,5,14)
hold on, plot(Ctrue(:,3),'k'), plot(C_j3(:,3),'b'),plot(C_p_j3(:,3),'r')
subplot(5,5,19)
hold on, plot(ctruemal,'k'), plot(C_p_j3(:,mal),'r')
subplot(5,5,24)
hold on, plot(ctruepro,'k'), plot(C_p_j3(:,pro),'r')
