function [A_aux,B_aux,C_aux,A_p_aux,B_p_aux,C_p_aux,err_als_vec]=     als_exact_c(Y,Y_p,iter,A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,sig1,sig2,r,compr)
%-------------------------------------------------------------------------%
% [A_aux,B_aux,C_aux,err_als_vec]=     als_exact(Y,Y_p,iter,A_0,B_0,C_0,A_p_0,B_p_0,C_p_0)
%
% Alternating least squares algorithm for 3-way CP decomposition. Factors A
% and B are normalized with the L_2 norm at each iterate. Enhanced to
% handle partial exact couplings though parametrization. IMPLEMENTED FOR
% NON-NEGATIVE DATA. COMPRESSED.
%
% Inputs (_p for second data block):
% 
% - Y           : data block (already compressed);
% - iter        : maximum number of iterates;
% - R           : number of components;
% - A_0,B_0,C_0 : initial parameters for the factors.
% - sig1,sig2   : noise levels of each tensor
% - r           : number of coupled columns. Have to be the r first columns
% of each exactly coupled factors.
% - compr       : compression matrices
%
% Outputs:
% 
% - A_aux       : first factor;
% - B_aux       : second factor;
% - C_aux       : third factor;
% - err_als_vec : iterations (Row 1) x reconstruction errors (Row 2).
%
% External functions:
%
% - kr.m        : Khatri-Rao product.
%
% List of updates                 -     03/01/2015  -     R. C. Farias
%                                       Creation of the file
%                                 -     20/07/2015  -     R. C. Farias
%                                       Updated with fast right inverse
%-------------------------------------------------------------------------%


%-------------------------- Compression things ---------------------------%
U = compr{1};
U_p = compr{4};
V = compr{2};
V_p = compr{5};
W = compr{3};

   A_0    =     (U')*A_0;
   B_0    =     (V')*B_0;
   C_0    =     (W')*C_0;
   A_p_0  =     (U_p')*A_p_0;         
   B_p_0  =     (V_p')*B_p_0;
   C_p_0  =     (W')*C_p_0;

%-------------------------------Parameters--------------------------------%
% Number of iterations for each evaluation of the reconstruction error
iter_err  =     10;
% Size of the filter for the relative error decrease
n_diff_err=     3;
% Stopping criteria parameters
stop_diff =     eps;
stop_err  =     eps;
% Size of the tensor
dim       =     size(Y);
dim_p     =     size(Y_p);
% Index for n-mode matricization
J         =     prod(dim)./dim;
J_p       =     prod(dim_p)./dim_p;
% Maximum number of iterations (/iter_err)
iter      =     ceil(iter/iter_err);
%-------------------------------------------------------------------------%

%-----------------------Algorithm initialization--------------------------%
% Three different unfoldings of the tensor
Y_mode    =     cell(3);
% Data tensor unfolding
Y_mode{1} =     reshape(permute(Y,[1,2,3]),dim(1),J(1));  
Y_mode{2} =     reshape(permute(Y,[2,1,3]),dim(2),J(2));
Y_mode{3} =     reshape(permute(Y,[3,1,2]),dim(3),J(3));
% Three different unfoldings of the tensor
Y_p_mode    =     cell(3);
% Data tensor unfolding
Y_p_mode{1} =     reshape(permute(Y_p,[1,2,3]),dim_p(1),J_p(1));  
Y_p_mode{2} =     reshape(permute(Y_p,[2,1,3]),dim_p(2),J_p(2));
Y_p_mode{3} =     reshape(permute(Y_p,[3,1,2]),dim_p(3),J_p(3));
% Storing auxiliary variables
A_aux     =     A_0;
B_aux     =     B_0;
C_aux     =     C_0;
A_p_aux   =     A_p_0;
B_p_aux   =     B_p_0;
C_p_aux   =     C_p_0;
% Filter for the relative error decrease
diff_err_filt   =     1000*ones(1,n_diff_err);
% Initial squared norm of the error
res             =     Y_mode{1}-(A_aux)*transpose(kr(C_aux,B_aux));
res_p           =     Y_p_mode{1}-(A_p_aux)*transpose(kr(C_p_aux,B_p_aux));
err_als_old     =     (res(:)')*res(:) + (res_p(:)')*res_p(:);
% Vector of squared norms of the error
err_als_vec     =     zeros(1,iter+1);
% First error
err_als_vec(1)  =     err_als_old;
fprintf('\nALS_Exact\n')
fprintf('----\n')
%-------------------------------------------------------------------------%

%--------------------------Algorithm iterate------------------------------%
t         =     0;
while     (mean(diff_err_filt)>stop_diff &&  t<iter ...
          &&  err_als_old>stop_err   )
    % Iteration counter (counts/iter_err)
    t=    t+1;
    % ANLS (iter_err iterates)
    for   i           =     1:iter_err
    
    % Least squares solution for each factor
    
    % A updates
    temp_CB           =     (C_aux'*C_aux).*(B_aux'*B_aux);
    A_aux             =     (Y_mode{1}*kr(C_aux,B_aux))/temp_CB;
    tempA             =     U*A_aux;
    tempA(tempA<0)    =     eps;
    A_aux             =     U'*tempA;
   % norm_A            =     sqrt(sum(A_aux.^2));
   % A_aux             =     A_aux.*repmat(1./norm_A,dim(1),1);
    
    temp_CB_p         =     (C_p_aux'*C_p_aux).*(B_p_aux'*B_p_aux);
    A_p_aux           =     (Y_p_mode{1}*kr(C_p_aux,B_p_aux))/temp_CB_p;
    tempA             =     U_p*A_p_aux;
    tempA(tempA<0)    =     eps;
    A_p_aux           =     U_p'*tempA;
   % norm_A_p          =     sqrt(sum(A_p_aux.^2));
   % A_p_aux           =     A_p_aux.*repmat(1./norm_A_p,dim_p(1),1);
    
    % B updates
    temp_CA           =     (C_aux'*C_aux).*(A_aux'*A_aux);
    B_aux             =     (Y_mode{2}*kr(C_aux,A_aux))/temp_CA;
    tempB             =     V*B_aux;
    tempB(tempB<0)    =     eps;
    B_aux             =     V'*tempB;
    norm_B            =     sqrt(sum(B_aux.^2));
    B_aux             =     B_aux.*repmat(1./norm_B,dim(2),1);
    
    temp_CA_p         =     (C_p_aux'*C_p_aux).*(A_p_aux'*A_p_aux);
    B_p_aux           =     (Y_p_mode{2}*kr(C_p_aux,A_p_aux))/temp_CA_p;
    tempB             =     V_p*B_p_aux;
    tempB(tempB<0)    =     eps;
    B_p_aux           =     V_p'*tempB;
    norm_B_p          =     sqrt(sum(B_p_aux.^2)); 
    B_p_aux           =     B_p_aux.*repmat(1./norm_B_p,dim_p(2),1);

    % C updates
    temp_BA_r         =     (B_aux(:,1:r)'*B_aux(:,1:r)).*(A_aux(:,1:r)'*A_aux(:,1:r));
    Y_BA_r            =     Y_mode{3}*kr(B_aux(:,1:r),A_aux(:,1:r))-...
                            C_aux(:,r+1:end)*((B_aux(:,r+1:end)'*B_aux(:,1:r)).*(A_aux(:,r+1:end)'*A_aux(:,1:r)));
    
    if (size(A_aux,2)> r)                    
    temp_BA           =     (B_aux(:,r+1:end)'*B_aux(:,r+1:end)).*(A_aux(:,r+1:end)'*A_aux(:,r+1:end));
    Y_BA              =     Y_mode{3}*kr(B_aux(:,r+1:end),A_aux(:,r+1:end))-...
                            C_aux(:,1:r)*((B_aux(:,1:r)'*B_aux(:,r+1:end)).*(A_aux(:,1:r)'*A_aux(:,r+1:end)));
    end
        
    temp_BA_p_r       =     (B_p_aux(:,1:r)'*B_p_aux(:,1:r)).*(A_p_aux(:,1:r)'*A_p_aux(:,1:r));
    Y_BA_p_r          =     Y_p_mode{3}*kr(B_p_aux(:,1:r),A_p_aux(:,1:r))-...
                            C_p_aux(:,r+1:end)*((B_p_aux(:,r+1:end)'*B_p_aux(:,1:r)).*(A_p_aux(:,r+1:end)'*A_p_aux(:,1:r)));
    
    if (size(A_p_aux,2)> r)                    
    temp_BA_p         =     (B_p_aux(:,r+1:end)'*B_p_aux(:,r+1:end)).*(A_p_aux(:,r+1:end)'*A_p_aux(:,r+1:end));
    Y_BA_p            =     Y_p_mode{3}*kr(B_p_aux(:,r+1:end),A_p_aux(:,r+1:end))-...
                            C_p_aux(:,1:r)*((B_p_aux(:,1:r)'*B_p_aux(:,r+1:end)).*(A_p_aux(:,1:r)'*A_p_aux(:,r+1:end)));
    end
    
    C_new_r           =     (Y_BA_r/sig1^2 + Y_BA_p_r/sig2^2)/(temp_BA_r/sig1^2+temp_BA_p_r/sig2^2);
    norm_c_new_r      =     sqrt(sum(C_new_r.^2));
    C_new_r           =     C_new_r.*repmat(1./norm_c_new_r,dim(3),1);
    if (size(A_aux,2)> r)                    
    C_new             =     Y_BA/temp_BA;
    norm_c_new        =     sqrt(sum(C_new.^2));
    C_new             =     C_new.*repmat(1./norm_c_new,dim(3),1);
    end
    if (size(A_p_aux,2)> r)                    
    C_p_new           =     Y_BA_p/temp_BA_p;
    norm_c_p_new      =     sqrt(sum(C_p_new.^2));
    C_p_new           =     C_p_new.*repmat(1./norm_c_p_new,dim_p(3),1);
    end
    
    if (size(A_aux,2)> r)                    
    C_aux             =     [C_new_r,C_new];
    else
    C_aux             =     C_new_r;
    end
    if (size(A_p_aux,2)> r)                    
    C_p_aux           =     [C_new_r,C_p_new];    
    else
    C_p_aux           =     C_new_r;
    end
    
    
    tempC             =     W*C_aux;
    tempC(tempC<0)    =     eps;
    C_aux             =     W'*tempC;
    
    tempC             =     W*C_p_aux;
    tempC(tempC<0)    =     eps;
    C_p_aux           =     W'*tempC;
    
    end
    % Squared norm of the error (it uses some precalculated quantities)
    err_als           =     sum(sum((Y_mode{3}-C_aux*kr(B_aux,A_aux)').^2))...
                           +sum(sum((Y_p_mode{3}-C_p_aux*kr(B_p_aux,A_p_aux)').^2));%snorm_Y-2*
    
    % Evaluation of the relative error decrease
    diff_err_filt(1:end-1)  =     diff_err_filt(2:end);
    diff_err_filt(end)      =     abs((err_als-err_als_old)/err_als);
    
    % Storing the last error
    err_als_old       =     err_als;
    err_als_vec(t+1)  =     err_als;
    
    if mod(t*iter_err,100)  ==    0
    % Printing the iterate error at each 100 iterates
    fprintf('error : %g  \tit : %d\n',err_als,t*iter_err)
    end
end


%----------------------- Ddecompression things ---------------------------%
A_aux = U*A_aux;
B_aux = V*B_aux;
C_aux = W*C_aux;
A_p_aux = U_p*A_p_aux;
B_p_aux = V_p*B_p_aux;
C_p_aux = W*C_p_aux;

A_aux(  A_aux<0)=0;
A_p_aux(A_p_aux<0)=0;
B_aux(  B_aux<0)=0;
B_p_aux(B_p_aux<0)=0;
C_aux(  C_aux<0)=0;
C_p_aux(C_p_aux<0)=0;

%-------------------------------------------------------------------------%
end