%---------------------------coupled_compression.m-------------------------%

%-------------------------------------------------------------------------%

%--------------Cleaning the workspace and loading functions---------------%
clc;
clear     all;
close     all;

%-------------------------------------------------------------------------%

%-----------------------------Model parameters----------------------------%
% Uncoumpressed tensor dimensions
% First tensor
dim_u     =     [21 251 28];
%dim_u     =     [21 100 28];
% Second tensor
dim_u_p   =     [8 13324 dim_u(3)];
%dim_u_p   =     [8 133 dim_u(3)];
% Number of components
R         =     [4 5];
% Number of coupled components
R_c       =     3;
% Compressed dimensions for uncoupled model
% First tensor
dim_c_u   =     [R(1) R(1) R(1)];
% Second tensor
dim_c_p_u =     [R(2) R(2) R(2)];
% Compressed dimensions for the coupled model
% First tensor
dim_c_c   =     [R(1) R(1) R(1)+R(2)-R_c];
% Second tensor
dim_c_p_c =     [R(2) R(2) R(2)+R(1)-R_c];
% Signal to noise ratio (SNR) for the tensors
SNR       =     [15 15];
% Standard deviation of the coupling
sigma_c   =     0.001;
% Number of uncoupled ALS iterates
iter_u    =     2000;
% Number of coupled ALS iterates
iter_c    =     2000;
% Number of initializations
n_init    =     10;
%-------------------------------------------------------------------------%

%-----------------------------Factors generation--------------------------%
% Standard deviations of the measurements noise
sigma_n   =     [sqrt((R(1)/(dim_u(1)*dim_u(2)))*10^(-SNR(1)/10)) ...
                sqrt((R(2)/(dim_u_p(1)*dim_u_p(2)))*10^(-SNR(2)/10))];
% Number of uncoupled components
R_u       =     R-[1 1]*R_c;
% Fixing the seed to have the same tensor
%rng(107)
% Random factors generation
A         =     abs(randn(dim_u(1),R(1)));
%A(:,1)    =     abs(0.6*A(:,4)+0.1*randn(dim_u(1),1));
B         =     abs(randn(dim_u(2),R(1)));
A_p       =     abs(randn(dim_u_p(1),R(2)));
B_p       =     abs(randn(dim_u_p(2),R(2)));
C_p       =     abs(randn(dim_u_p(3),R(2)));
% Normalization
A         =     A.*repmat(1./sqrt(sum(A.^2)),dim_u(1),1);
B         =     B.*repmat(1./sqrt(sum(B.^2)),dim_u(2),1);
A_p       =     A_p.*repmat(1./sqrt(sum(A_p.^2)),dim_u_p(1),1);
B_p       =     B_p.*repmat(1./sqrt(sum(B_p.^2)),dim_u_p(2),1);
% Noise on the coupling
% Coupled components
C(:,1:R_c)      =     abs(C_p(:,1:R_c)+sigma_c*randn(dim_u(3),R_c));
% Uncoupled components
C(:,R_c+1:R(1)) =     abs(randn(dim_u(3),R_u(1)));
% First noiseless tensor
X         =     reshape(A*transpose(kr(C,B)),dim_u);
% Second noiseless tensor
X_p       =     reshape(A_p*transpose(kr(C_p,B_p)),dim_u_p);
%-------------------------------------------------------------------------%

%---------------------------------Coupling model--------------------------%
% Coupling model in the uncompressed domain
% Constraint matrices for vectorized factors
% First tensor
H_u       =     [eye(dim_u(3)*R_c) zeros([dim_u(3)*R_c dim_u(3)*R_u(1)])];
% Second tensor
H_p_u     =     [eye(dim_u_p(3)*R_c) ...
                              zeros([dim_u_p(3)*R_c dim_u_p(3)*(R_u(2))])];

% Coupling model in the compressed domain
% Constraint matrices for vectorized factors
% First tensor
H_c       =     [eye(dim_c_c(3)*R_c) ...
                                zeros([dim_c_c(3)*R_c dim_c_c(3)*R_u(1)])];
% Second tensor
H_p_c     =     [eye(dim_c_p_c(3)*R_c) ...
                          zeros([dim_c_p_c(3)*R_c dim_c_p_c(3)*(R_u(2))])];
%-------------------------------------------------------------------------%

%------------------------Noisy tensors construction-----------------------%
% Noisy measurements
% First noisy tensor
Y   =     X+sigma_n(1)*randn(dim_u);
Y(Y<0)    =     0;
% Second noisy tensor
Y_p =     X_p+sigma_n(2)*randn(dim_u_p);
Y_p(Y_p<0)      =     0;
%-------------------------------------------------------------------------%

%---------------------Uncoupled tensor compression------------------------%
type=     'standard';
% First tensor
[~,~,~,~,~,~,S_u]   =     hosvd2(Y,dim_c_u,type);
% Second tensor
[~,~,~,~,~,~,S_p_u] =     hosvd2(Y_p,dim_c_p_u,type);
%-------------------------------------------------------------------------%

%-----------------------Coupled tensor compression------------------------%
type=     'standard';
[~,~,~,~,S_c,~,~,~,~,S_p_c] =  ...
                       coupled_hosvd(Y,Y_p,dim_c_c,dim_c_p_c,sigma_n,type);
%-------------------------------------------------------------------------%

%----------------------------Results Fig. 1-------------------------------%
figure
subplot(311)
plot(diag(S_u))
subplot(312)
plot(diag(S_p_u))
subplot(313)
plot(S_c(3,:))
%-------------------------------------------------------------------------%

%--------Different initializations for the uncompressed approach----------%
% Initial cost function values for picking best initialization
obj_n_val_u_1   =     10^50;
obj_n_val_u_2   =     10^50;
map_n_val_u     =     10^50;
for i     =     1:n_init
%-----------------------------Initialization------------------------------%
A_0       =     abs(randn([dim_u(1) R(1)]));
B_0       =     abs(randn([dim_u(2) R(1)]));
A_0       =     A_0.*repmat(1./sqrt(sum(A_0.^2)),dim_u(1),1);
B_0       =     B_0.*repmat(1./sqrt(sum(B_0.^2)),dim_u(2),1);
   
A_p_0     =     abs(randn([dim_u_p(1) R(2)]));
B_p_0     =     abs(randn([dim_u_p(2) R(2)]));
A_p_0     =     A_p_0.*repmat(1./sqrt(sum(A_p_0.^2)),dim_u_p(1),1);
B_p_0     =     B_p_0.*repmat(1./sqrt(sum(B_p_0.^2)),dim_u_p(2),1);
 
C_0       =     abs(randn([dim_u(3) R(1)]));
C_p_0     =     abs(randn([dim_u_p(3) R(2)]));
%-------------------------------------------------------------------------%

%-----------------Algorithms in the uncompressed domain-------------------%
% Option for nonnegative ALS
option    =     'NN';  
%-----------------------------Uncoupled CP ALS----------------------------%
% ALS for the first tensor
[A_n_u_u,B_n_u_u,C_n_u_u,obj_val_vec_1] =     anls(Y,iter_u,A_0,B_0,C_0);
% ALS for the second tensor
[A_p_n_u_u,B_p_n_u_u,C_p_n_u_u,obj_val_vec_2] = ...
                                        anls(Y_p,iter_u,A_p_0,B_p_0,C_p_0);
[A_n_u_u,B_n_u_u,C_n_u_u,A_p_n_u_u,B_p_n_u_u,C_p_n_u_u]            =    ...
                                 amb_correct_1f(A_n_u_u,B_n_u_u,C_n_u_u,...
                                            A_p_n_u_u,B_p_n_u_u,C_p_n_u_u);
if obj_val_vec_1(end)<obj_n_val_u_1
obj_n_val_u_1   =     obj_val_vec_1(end);
A_u_u           =     A_n_u_u;
B_u_u           =     B_n_u_u;
C_u_u           =     C_n_u_u;
end
if obj_val_vec_2(end)<obj_n_val_u_2
obj_n_val_u_2   =     obj_val_vec_2(end);
A_p_u_u         =     A_p_n_u_u;
B_p_u_u         =     B_p_n_u_u;
C_p_u_u         =     C_p_n_u_u;
end
%-------------------------------------------------------------------------%

%-----------------------Flexibly coupled CP ALS---------------------------%
A_0       =     A_n_u_u;
B_0       =     B_n_u_u;
C_0       =     C_n_u_u;
A_p_0     =     A_p_n_u_u;
B_p_0     =     B_p_n_u_u;
C_p_0     =     C_p_n_u_u;
% CCP_ALS
[A_n_c_u,B_n_c_u,C_n_c_u,A_p_n_c_u,B_p_n_c_u,C_p_n_c_u,map_val_vec] = ...
              ccp_als(Y,Y_p,iter_c,H_u,H_p_u,sigma_n,sigma_c,'joint',...
                                 A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,option,0,0);

if map_val_vec(end)<map_n_val_u
map_n_val_u     =     map_val_vec(end);
A_c_u     =     A_n_c_u;
B_c_u     =     B_n_c_u;
C_c_u     =     C_n_c_u;

A_p_c_u   =     A_p_n_c_u;
B_p_c_u   =     B_p_n_c_u;
C_p_c_u   =     C_p_n_c_u;
end
%-------------------------------------------------------------------------%

end
%-------------------------------------------------------------------------%

%----------Different initializations for the compressed approach----------%

%-----------------------Coupled tensor compression------------------------%
type=     'standard';
[G_c,U_c,V_c,W_c,~,G_p_c,U_p_c,V_p_c,W_p_c,~] =     ...
                       coupled_hosvd(Y,Y_p,dim_c_c,dim_c_p_c,sigma_n,type);
%-------------------------------------------------------------------------%

% Initial cost function values for picking best initialization
obj_n_val_c_1   =     10^50;
obj_n_val_c_2   =     10^50;
map_n_val_c     =     10^50;
for i     =     1:n_init
%-----------------------------Initialization------------------------------%
A_0       =     abs(randn([dim_u(1) R(1)]));
B_0       =     abs(randn([dim_u(2) R(1)]));
A_0       =     A_0.*repmat(1./sqrt(sum(A_0.^2)),dim_u(1),1);
B_0       =     B_0.*repmat(1./sqrt(sum(B_0.^2)),dim_u(2),1);
   
A_p_0     =     abs(randn([dim_u_p(1) R(2)]));
B_p_0     =     abs(randn([dim_u_p(2) R(2)]));
A_p_0     =     A_p_0.*repmat(1./sqrt(sum(A_p_0.^2)),dim_u_p(1),1);
B_p_0     =     B_p_0.*repmat(1./sqrt(sum(B_p_0.^2)),dim_u_p(2),1);
 
C_0       =     abs(randn([dim_u(3) R(1)]));
C_p_0     =     abs(randn([dim_u_p(3) R(2)]));
%-------------------------------------------------------------------------%

%------------------Algorithms in the compressed domain--------------------%
% Option for nonnegative ALS
option_1  =     'NN';
% Option for ALS in the compressed domain
option_2  =     'yes';
%------------------------------PROCO ALS----------------------------------%
% PROCO-ALS for the first tensor
[A_n_u_c,B_n_u_c,C_n_u_c,obj_val_vec_1] =    proco_als(0,iter_u,dim_c_c,...
                                                     'standard','given',...
                                                            A_0,B_0,C_0,...
                                                        0,G_c,U_c,V_c,W_c);
% PROCO-ALS for the second tensor
[A_p_n_u_c,B_p_n_u_c,C_p_n_u_c,obj_val_vec_2] = proco_als(0,iter_u,...
                                             dim_c_c,'standard','given',...
                                                      A_p_0,B_p_0,C_p_0,...
                                                0,G_p_c,U_p_c,V_p_c,W_p_c);
[A_n_u_c,B_n_u_c,C_n_u_c,A_p_n_u_c,B_p_n_u_c,C_p_n_u_c]            =    ...
                                 amb_correct_1f(A_n_u_c,B_n_u_c,C_n_u_c,...
                                            A_p_n_u_c,B_p_n_u_c,C_p_n_u_c);
if obj_val_vec_1(end)<obj_n_val_c_1
obj_n_val_c_1   =     obj_val_vec_1(end);
A_u_c           =     A_n_u_c;
B_u_c           =     B_n_u_c;
C_u_c           =     C_n_u_c;
end
if obj_val_vec_2(end)<obj_n_val_c_2
obj_n_val_c_2   =     obj_val_vec_2(end);
A_p_u_c         =     A_p_n_u_c;
B_p_u_c         =     B_p_n_u_c;
C_p_u_c         =     C_p_n_u_c;
end
%-------------------------------------------------------------------------%

%-----------------------Flexibly coupled CP ALS---------------------------%
A_0       =     A_n_u_c;
B_0       =     B_n_u_c;
C_0       =     C_n_u_c;
A_p_0     =     A_p_n_u_c;
B_p_0     =     B_p_n_u_c;
C_p_0     =     C_p_n_u_c;
% Cell structure with compression matrices
M{1}      =     U_c;
M{2}      =     V_c;
M{3}      =     W_c;
M{4}      =     U_p_c;
M{5}      =     V_p_c;
% CCP_ALS
[A_n_c_c,B_n_c_c,C_n_c_c,A_p_n_c_c,B_p_n_c_c,C_p_n_c_c,map_val_vec] = ...
          ccp_als(G_c,G_p_c,iter_c,H_c,H_p_c,sigma_n,sigma_c,'joint',...
                        A_0,B_0,C_0,A_p_0,B_p_0,C_p_0,option_1,option_2,M);

if map_val_vec(end)<map_n_val_u
map_n_val_u     =     map_val_vec(end);
A_c_c     =     A_n_c_c;
B_c_c     =     B_n_c_c;
C_c_c     =     C_n_c_c;

A_p_c_c   =     A_p_n_c_c;
B_p_c_c   =     B_p_n_c_c;
C_p_c_c   =     C_p_n_c_c;
end
%-------------------------------------------------------------------------%

end
%-------------------------------------------------------------------------%


%---------------------------- Normalizations -----------------------------%

C               =     C.*repmat(sqrt(sum(A.^2)).*sqrt(sum(B.^2)),dim_u(3),1);
A               =     A.*repmat(1./sqrt(sum(A.^2)),dim_u(1),1);
B               =     B.*repmat(1./sqrt(sum(B.^2)),dim_u(2),1);


C_p             =     C_p.*repmat(sqrt(sum(A_p.^2)).*sqrt(sum(B_p.^2)),dim_u_p(3),1);
A_p             =     A_p.*repmat(1./sqrt(sum(A_p.^2)),dim_u_p(1),1);
B_p             =     B_p.*repmat(1./sqrt(sum(B_p.^2)),dim_u_p(2),1);


C_u_u           =     C_u_u.*repmat(sqrt(sum(A_u_u.^2)).*sqrt(sum(B_u_u.^2)),dim_u(3),1);
A_u_u           =     A_u_u.*repmat(1./sqrt(sum(A_u_u.^2)),dim_u(1),1);
B_u_u           =     B_u_u.*repmat(1./sqrt(sum(B_u_u.^2)),dim_u(2),1);


C_p_u_u         =     C_p_u_u.*repmat(sqrt(sum(A_p_u_u.^2)).*sqrt(sum(B_p_u_u.^2)),dim_u_p(3),1);
A_p_u_u         =     A_p_u_u.*repmat(1./sqrt(sum(A_p_u_u.^2)),dim_u_p(1),1);
B_p_u_u         =     B_p_u_u.*repmat(1./sqrt(sum(B_p_u_u.^2)),dim_u_p(2),1);


C_c_u           =     C_c_u.*repmat(sqrt(sum(A_c_u.^2)).*sqrt(sum(B_c_u.^2)),dim_u(3),1);
A_c_u           =     A_c_u.*repmat(1./sqrt(sum(A_c_u.^2)),dim_u(1),1);
B_c_u           =     B_c_u.*repmat(1./sqrt(sum(B_c_u.^2)),dim_u(2),1);


C_p_c_u         =     C_p_c_u.*repmat(sqrt(sum(A_p_c_u.^2)).*sqrt(sum(B_p_c_u.^2)),dim_u_p(3),1);
A_p_c_u         =     A_p_c_u.*repmat(1./sqrt(sum(A_p_c_u.^2)),dim_u_p(1),1);
B_p_c_u         =     B_p_c_u.*repmat(1./sqrt(sum(B_p_c_u.^2)),dim_u_p(2),1);


C_u_c           =     C_u_c.*repmat(sqrt(sum(A_u_c.^2)).*sqrt(sum(B_u_c.^2)),dim_u(3),1);
A_u_c           =     A_u_c.*repmat(1./sqrt(sum(A_u_c.^2)),dim_u(1),1);
B_u_c           =     B_u_c.*repmat(1./sqrt(sum(B_u_c.^2)),dim_u(2),1);


C_p_u_c         =     C_p_u_c.*repmat(sqrt(sum(A_p_u_c.^2)).*sqrt(sum(B_p_u_c.^2)),dim_u_p(3),1);
A_p_u_c         =     A_p_u_c.*repmat(1./sqrt(sum(A_p_u_c.^2)),dim_u_p(1),1);
B_p_u_c         =     B_p_u_c.*repmat(1./sqrt(sum(B_p_u_c.^2)),dim_u_p(2),1);


C_c_c           =     C_u_c.*repmat(sqrt(sum(A_u_c.^2)).*sqrt(sum(B_u_c.^2)),dim_u(3),1);
A_c_c           =     A_u_c.*repmat(1./sqrt(sum(A_u_c.^2)),dim_u(1),1);
B_c_c           =     B_u_c.*repmat(1./sqrt(sum(B_u_c.^2)),dim_u(2),1);

C_p_c_c         =     C_p_c_c.*repmat(sqrt(sum(A_p_c_c.^2)).*sqrt(sum(B_p_c_c.^2)),dim_u_p(3),1);
A_p_c_c         =     A_p_c_c.*repmat(1./sqrt(sum(A_p_c_c.^2)),dim_u_p(1),1);
B_p_c_c         =     B_p_c_c.*repmat(1./sqrt(sum(B_p_c_c.^2)),dim_u_p(2),1);



%--------Correcting the ambiguity and evaluating the squared error--------%
% Uncoupled uncompressed CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_u_u,B_u_u,C_u_u]   =     amb_correct(A_u_u,B_u_u,C_u_u,A,B,C);

% Correcting permutation and sign ambiguity for the second tensor
[A_p_u_u,B_p_u_u,C_p_u_u]   =...
                      amb_correct(A_p_u_u,B_p_u_u,C_p_u_u,A_p,B_p,C_p);
% Coupled uncompressed CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_c_u,B_c_u,C_c_u]   =     amb_correct(A_c_u,B_c_u,C_c_u,A,B,C);

% Correcting permutation and sign ambiguity for the second tensor
[A_p_c_u,B_p_c_u,C_p_c_u]   =...
                      amb_correct(A_p_c_u,B_p_c_u,C_p_c_u,A_p,B_p,C_p);
% Uncoupled compressed CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_u_c,B_u_c,C_u_c]   =     amb_correct(A_u_c,B_u_c,C_u_c,A,B,C);

% Correcting permutation and sign ambiguity for the second tensor
[A_p_u_c,B_p_u_c,C_p_u_c]   =...
                      amb_correct(A_p_u_c,B_p_u_c,C_p_u_c,A_p,B_p,C_p);
% Coupled compressed CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_c_c,B_c_c,C_c_c]   =     amb_correct(A_c_c,B_c_c,C_c_c,A,B,C);

% Correcting permutation and sign ambiguity for the second tensor
[A_p_c_c,B_p_c_c,C_p_c_c]   =...
                      amb_correct(A_p_c_c,B_p_c_c,C_p_c_c,A_p,B_p,C_p);
%-------------------------------------------------------------------------%

%---------------------------Total squared errors--------------------------%
% Uncoupled uncompressed
tse_A_u_u       =     sum((A(:)-A_u_u(:)).^2);
tse_B_u_u       =     sum((B(:)-B_u_u(:)).^2);
tse_C_u_u       =     sum((C(:)-C_u_u(:)).^2);
tse_A_p_u_u     =     sum((A_p(:)-A_p_u_u(:)).^2);
tse_B_p_u_u     =     sum((B_p(:)-B_p_u_u(:)).^2);
tse_C_p_u_u     =     sum((C_p(:)-C_p_u_u(:)).^2);
% Coupled uncompressed
tse_A_c_u       =     sum((A(:)-A_c_u(:)).^2);
tse_B_c_u       =     sum((B(:)-B_c_u(:)).^2);
tse_C_c_u       =     sum((C(:)-C_c_u(:)).^2);
tse_A_p_c_u     =     sum((A_p(:)-A_p_c_u(:)).^2);
tse_B_p_c_u     =     sum((B_p(:)-B_p_c_u(:)).^2);
tse_C_p_c_u     =     sum((C_p(:)-C_p_c_u(:)).^2);
% Uncoupled compressed
tse_A_u_c       =     sum((A(:)-A_u_c(:)).^2);
tse_B_u_c       =     sum((B(:)-B_u_c(:)).^2);
tse_C_u_c       =     sum((C(:)-C_u_c(:)).^2);
tse_A_p_u_c     =     sum((A_p(:)-A_p_u_c(:)).^2);
tse_B_p_u_c     =     sum((B_p(:)-B_p_u_c(:)).^2);
tse_C_p_u_c     =     sum((C_p(:)-C_p_u_c(:)).^2);
% Coupled compressed
tse_A_c_c       =     sum((A(:)-A_c_c(:)).^2);
tse_B_c_c       =     sum((B(:)-B_c_c(:)).^2);
tse_C_c_c       =     sum((C(:)-C_c_c(:)).^2);
tse_A_p_c_c     =     sum((A_p(:)-A_p_c_c(:)).^2);
tse_B_p_c_c     =     sum((B_p(:)-B_p_c_c(:)).^2);
tse_C_p_c_c     =     sum((C_p(:)-C_p_c_c(:)).^2);
%-------------------------------------------------------------------------%

%----------------------------Results Tab. 1-------------------------------%
fprintf('Alg. \t \t A \t \t B \t \t C \t \t A_p \t \t B_p \t \t C_p\n')
fprintf('Unc. unc. \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f\n',...
         tse_A_u_u,tse_B_u_u,tse_C_u_u,tse_A_p_u_u,tse_B_p_u_u,tse_C_p_u_u)
fprintf('Coup. unc. \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f\n',...
         tse_A_c_u,tse_B_c_u,tse_C_c_u,tse_A_p_c_u,tse_B_p_c_u,tse_C_p_c_u)
fprintf('Unc. comp. \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f\n',...
         tse_A_u_c,tse_B_u_c,tse_C_u_c,tse_A_p_u_c,tse_B_p_u_c,tse_C_p_u_c)
fprintf('Coup. comp. \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f \t %.5f\n',...
         tse_A_c_c,tse_B_c_c,tse_C_c_c,tse_A_p_c_c,tse_B_p_c_c,tse_C_p_c_c)
%-------------------------------------------------------------------------%