function [G_1,U_c_1,V_c_1,W_c_1,S_1,G_2,U_c_2,V_c_2,W_c_2,S_2]    =  ...
                        coupled_hosvd(Y_1,Y_2,dim_c_1,dim_c_2,sigma_n,type)
%-------------------------------------------------------------------------%
% [G_1,U_c_1,V_c_1,W_c_1,G_2,U_c_2,V_c_2,W_c_2]    =     ...
%                       coupled_hosvd(Y_1,Y_2,dim_c_1,dim_c_2,sigma_n,type)
%
% This function approximates the coupled HOSVD of tensors Y_1 and Y_2 with
% approximations dim_c_1 and dim_c_2 for ranks of unfolding matrices
% (dimensions of the core tensor). SVD can be evaluated by the standard 
% method or by the randomized approximation. Since the third mode is
% coupled truncated SVD is carried out jointly in this mode, by stacking
% the weighted third unfoldings of the tensors.
% 
% Inputs:
% 
% - Y_1, Y_2    : data blocks;
% - dim_c_1,... : dimensions of the compressed spaces;
% - sigma_n     : noise standard deviations;
% - type        : type of method to be used to evaluate the SVD.
%
%                 For type='standard', the standard method is used.
%
%                 For type='random', the randomized SVD is used.
%
% Outputs:
%
% - G_1, G_2    : core tensors (compressed data block);
% - S_1, S_2    : singular values of the uncoupled modes, each row
%                 corresponds to one unfolding and the third row 
%                 corresponds to the joint SVD;
% - U_c_1,...   : HOSVD factors (compression matrices).
%
% External functions:
%
% - rsvd.m      : Randomized SVD.
%
% List of updates                 -     21/07/2015  -     J. E. Cohen
%                                       Creation of the file
%                                 -     05/02/2016  -     R. C. Farias
%                                       Modification to deal with coupled
%                                       tensors
%-------------------------------------------------------------------------%

%-------------------------------Parameters--------------------------------%
% Size of the tensor
dim_u_1   =     size(Y_1);
dim_u_2   =     size(Y_2);
% Index for n-mode matricization
prod_u_1  =     prod(dim_u_1)./dim_u_1;
prod_u_2  =     prod(dim_u_2)./dim_u_2;
% Coupled dimension
dim_coup  =     dim_u_1(3)+dim_u_2(3);
% Maximum dimension
max_dim_1 =     max([dim_u_1 dim_coup]);
max_dim_2 =     max([dim_u_2 dim_coup]);
%-------------------------------------------------------------------------%

%--------------------Singular Value Decompositions------------------------%
% Allocation of singular values matrices
S_1       =     zeros(3,max_dim_1);
S_2       =     zeros(3,max_dim_2);
switch(type)

%---------------------------Standard SVD----------------------------------%
    case('standard')
% Standard economy size SVD of the unfoldings
% SVD of the coupled mode
% Unfolding (3)
Y_unf_1   =     reshape(permute(Y_1,[3,1,2]),dim_u_1(3),prod_u_1(3));
Y_unf_2   =     reshape(permute(Y_2,[3,1,2]),dim_u_2(3),prod_u_2(3));
[W,S,~]   =     svd([(1./sigma_n(1))*Y_unf_1 (1./sigma_n(2))*Y_unf_2]...
                ,'econ');
S_1(3,1:length(S))    =     diag(S);
S_2(3,1:length(S))    =     diag(S);
% Standard SVD of the unfoldings to obtain the left singular vectors
% Unfolding (2)
% First tensor
Y_unf_1   =     reshape(permute(Y_1,[2,1,3]),dim_u_1(2),prod_u_1(2));
[V_1,S,~]       =     svd(Y_unf_1,'econ');
S_1(2,1:length(S))    =     diag(S);
% Second tensor
Y_unf_2   =     reshape(permute(Y_2,[2,1,3]),dim_u_2(2),prod_u_2(2));
[V_2,S,~]       =     svd(Y_unf_2,'econ');
S_2(2,1:length(S))    =     diag(S);
% Unfolding (1)
% First tensor
Y_unf_1   =     reshape(Y_1,dim_u_1(1),prod_u_1(1));
[U_1,S,~] =     svd(Y_unf_1,'econ');
S_1(1,1:length(S))    =     diag(S);
% Second tensor
Y_unf_2   =     reshape(Y_2,dim_u_2(1),prod_u_2(1));
[U_2,S,~] =     svd(Y_unf_2,'econ');
S_2(1,1:length(S))    =     diag(S);
% Truncation of the left singular vectors (low rank approximation)
% First tensor
U_c_1     =     U_1(:,1:dim_c_1(1));
V_c_1     =     V_1(:,1:dim_c_1(2));
W_c_1     =     W(:,1:dim_c_1(3));
% Second tensor
U_c_2     =     U_2(:,1:dim_c_2(1));
V_c_2     =     V_2(:,1:dim_c_2(2));
W_c_2     =     W(:,1:dim_c_2(3));
%-------------------------------------------------------------------------%

%---------------------------Randomized SVD--------------------------------%
    case('random')
% RSVD of the unfoldings (one power iteration, result is already truncated)
% RSVD of the coupled mode
% Unfolding (3)
Y_unf_1   =     reshape(permute(Y_1,[3,1,2]),dim_u_1(3),prod_u_1(3));
Y_unf_2   =     reshape(permute(Y_2,[3,1,2]),dim_u_2(3),prod_u_2(3));
[W_c_1,S,~]     =     rsvd([(1./sigma_n(1))*Y_unf_1 ...
                                    (1./sigma_n(2))*Y_unf_2],dim_c_1(3),1);
W_c_2     =     W_c_1;
S_1(3,1:length(S))    =     diag(S);
S_2(3,1:length(S))    =     diag(S);
% Unfolding (2)
% First tensor
Y_unf_1   =     reshape(permute(Y_1,[2,1,3]),dim_u_1(2),prod_u_1(2));
[V_c_1,S,~]     =     rsvd(Y_unf_1,dim_c_1(2),1);
S_1(2,1:length(S))    =     diag(S);
% Second tensor
Y_unf_2   =     reshape(permute(Y_2,[2,1,3]),dim_u_2(2),prod_u_2(2));
[V_c_2,S,~]     =     rsvd(Y_unf_2,dim_c_2(2),1);
S_2(2,1:length(S))    =     diag(S);
% Unfolding (1)
% First tensor
Y_unf_1         =     reshape(Y_1,dim_u_1(1),prod_u_1(1));
[U_c_1,S,~]     =     rsvd(Y_unf_1,dim_c_1(1),1);
S_2(1,1:length(S))     =     diag(S);
% Second tensor
Y_unf_2         =     reshape(Y_2,dim_u_2(1),prod_u_2(1));
[U_c_2,S,~]     =     rsvd(Y_unf_2,dim_c_2(1),1);
S_2(1,1:length(S))    =     diag(S);
%-------------------------------------------------------------------------%
otherwise
    disp('Unknown method. Choose between: standard or random.')
end

% Evaluation of the core tensor
% First tensor
G_1 =     reshape(U_c_1'*Y_unf_1,dim_c_1(1),dim_u_1(2),dim_u_1(3));
clear     Y_unf_1;
T_2 =     reshape(permute(G_1,[2,1,3]),dim_u_1(2),dim_c_1(1)*dim_u_1(3));
G_1 =     permute(reshape(V_c_1'*T_2,dim_c_1(2),dim_c_1(1),dim_u_1(3)),...
                                                                  [2,1,3]);
T_3 =     reshape(permute(G_1,[3,1,2]),dim_u_1(3),dim_c_1(1)*dim_c_1(2));
G_1 =     permute(reshape(W_c_1'*T_3,dim_c_1(3),dim_c_1(1),dim_c_1(2)),...
                                                                  [2,3,1]);
                                                              
clear     T_2 T_3;
% First tensor
G_2 =     reshape(U_c_2'*Y_unf_2,dim_c_2(1),dim_u_2(2),dim_u_2(3));
clear     Y_unf_2;
T_2 =     reshape(permute(G_2,[2,1,3]),dim_u_2(2),dim_c_2(1)*dim_u_2(3));
G_2 =     permute(reshape(V_c_2'*T_2,dim_c_2(2),dim_c_2(1),dim_u_2(3)),...
                                                                  [2,1,3]);
T_3 =     reshape(permute(G_2,[3,1,2]),dim_u_2(3),dim_c_2(1)*dim_c_2(2));
G_2 =     permute(reshape(W_c_2'*T_3,dim_c_2(3),dim_c_2(1),dim_c_2(2)),...
                                                                  [2,3,1]);
%-------------------------------------------------------------------------%
end

